/*------------------------------------------------------------------------------
 * MDK Middleware - Component ::Network
 * Copyright (c) 2004-2017 ARM Germany GmbH. All rights reserved.
 *------------------------------------------------------------------------------
 * Name:    TLS_mbed.c
 * Purpose: TLS Interface for mbedTLS
 * Rev.:    V7.4.1
 *----------------------------------------------------------------------------*/

#include <string.h>
#include "rl_net_lib.h"

#include "mbedtls/config.h"
#include "mbedtls/debug.h"
#include "mbedtls/ssl.h"
#include "mbedtls/entropy.h"
#include "mbedtls/ctr_drbg.h"
#include "mbedtls/certs.h"
#include "mbedtls/entropy_poll.h"
#include "mbedtls/net.h"
#if defined(MBEDTLS_SSL_CACHE_C)
#include "mbedtls/ssl_cache.h"
#endif
#if defined(MBEDTLS_DEBUG_C)
#include <stdio.h>
#endif

#ifndef TLS_THREAD_STACK_SIZE
#define TLS_THREAD_STACK_SIZE   4096
#endif

#ifndef TLS_THREAD_PRIORITY
#define TLS_THREAD_PRIORITY     osPriorityNormal
#endif

/* mbedTLS debug level */
#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL             1
#endif

#if defined(RTE_Network_Web_Server_RO_TLS) || defined(RTE_Network_Web_Server_FS_TLS)
#include "..\Network\Net_Config_HTTP_Server.h"
#else
#define HTTP_SERVER_NUM_SESSIONS 0
#endif

#define NET_BUFFER_HLEN         8       // Size of NET_BUFFER header
#define QUEUE_HEADER_LEN        ((NET_BUFFER_HLEN - NET_HEADER_LEN) | 0x40000000)
#define TLS_NUM_SOCKS           (HTTP_SERVER_NUM_SESSIONS)

/* TLS states */
typedef enum {
  TLS_STATE_FREE = 0,                   // Entry is free and unused
  TLS_STATE_IDLE,                       // Idle, but allocated
  TLS_STATE_SETUP,                      // Setup session, next is handshaking
  TLS_STATE_HANDSHAKE,                  // Handshaking in progress
  TLS_STATE_APPDATA,                    // Encrypted data exchange
  TLS_STATE_CLOSE,                      // TLS session is closing
} TLS_STATE;

/* TLS control info */
typedef struct net_tls_info {
  TLS_STATE    State;                   // Current state
  uint8_t      Socket;                  // Assigned TCP socket
  bool         CloseRq;                 // Close requested by service
  netTCP_cb_t  cb_func;                 // Application callback function
  __ADDR       Peer;                    // Peer IP address and port
  int32_t      ssl_error;               // Fail error code
  void        *ssl;                     // SSL context
  NET_BUFFER  *t_buf;                   // Transmit data packet
  NET_BUFFER  *r_queue;                 // Receive data queue
} TLS_INFO;

/* Certificates */
extern const uint8_t  NetSecurity_ServerCA[];
extern const uint32_t NetSecurity_ServerCA_Len;
extern const uint8_t  NetSecurity_ServerCert[];
extern const uint32_t NetSecurity_ServerCert_Len;
extern const uint8_t  NetSecurity_ServerKey[];
extern const uint32_t NetSecurity_ServerKey_Len;

/* Local variables */
static struct {
  NETOS_ID thread_id;                   // TLS thread identifier
  int32_t  ssl_error;                   // tls_init() error code
  bool     init_done;                   // TLS layer initialized
  bool     running;                     // TLS thread currently running
  bool     busy;                        // need another call to tls_run()
} ctrl;
static TLS_INFO sec_scb[TLS_NUM_SOCKS];
static uint8_t sec_buf[1460];

/* mbedTLS resources */
static mbedtls_entropy_context entropy;
static mbedtls_ctr_drbg_context ctr_drbg;
static mbedtls_ssl_config conf;
static mbedtls_x509_crt srvcert;
static mbedtls_pk_context pkey;
static const char pers[] = "ssl_server";
#if defined(MBEDTLS_SSL_CACHE_C)
static mbedtls_ssl_cache_context cache;
#endif
static mbedtls_ssl_context ssl[TLS_NUM_SOCKS];

/* Local Functions */
static void netTLS_Thread (void const *arg) __attribute__((noreturn));
static int32_t tls_init (void);
static void    tls_run (void);
static void    thread_wakeup (void);
static TLS_INFO *tls_map_socket (int32_t socket);
static uint32_t cb_tls (int32_t socket, netTCP_Event event, const NET_ADDR *addr,
                                        const uint8_t *buf, uint32_t len);
static int32_t bio_recv (void *ctx, unsigned char *buf, uint32_t len);
static int32_t bio_send (void *ctx, const uint8_t *buf, uint32_t len);
static void queue_add_buf (TLS_INFO *sec_s, NET_BUFFER *netbuf);
static void queue_release (TLS_INFO *sec_s);
static void free_tx_buf (TLS_INFO *sec_s);

/* RTOS resources */
#if defined(RTE_CMSIS_RTOS)
  #include "cmsis_os.h"
  extern const osThreadDef_t os_thread_def_netTLS_Thread;
  osThreadDef(netTLS_Thread, TLS_THREAD_PRIORITY, 1, TLS_THREAD_STACK_SIZE);
  #define OS_THREAD_NEW(func)   osThreadCreate (osThread(func), NULL)
  #define OS_FLAG_WAIT(flag)    osSignalWait (flag, osWaitForever)
  #define OS_FLAG_SET(flag)     osSignalSet (ctrl.thread_id, flag)
#elif defined(RTE_CMSIS_RTOS2)
  #include "cmsis_os2.h"
  #define OS_THREAD_NEW(func)   osThreadNew ((osThreadFunc_t)&func, NULL, &tls_thread);
  #define OS_FLAG_WAIT(flag)    osThreadFlagsWait (flag, osFlagsWaitAny, osWaitForever)
  #define OS_FLAG_SET(flag)     osThreadFlagsSet (ctrl.thread_id, flag)
  #ifdef RTE_CMSIS_RTOS2_RTX5
    #include "rtx_os.h"
    static osRtxThread_t        tls_thread_cb;
    static uint64_t             tls_stack[TLS_THREAD_STACK_SIZE/8];
    #define __TLS_TCB           &tls_thread_cb, sizeof(tls_thread_cb)
    #define __TLS_STACK         &tls_stack,     sizeof(tls_stack)
  #else
    #define __TLS_TCB           NULL, 0
    #define __TLS_STACK         NULL, TLS_THREAD_STACK_SIZE
  #endif

  static const osThreadAttr_t tls_thread = {
    "netTLS_Thread",
    osThreadDetached,
    __TLS_TCB,
    __TLS_STACK,
    TLS_THREAD_PRIORITY,
    0,
    0
  };
#else
  #error "::CMSIS:RTOS selection invalid"
#endif

/**
  \brief       mbedTLS debug output.
  \return      none.
*/
#if defined(MBEDTLS_DEBUG_C)
static void out_debug (void *ctx, int level, const char *file,
                                             int line, const char *str ) {
#ifndef DEBUG_FULL_PATH
  /* Skip path ie. "C:\Keil\ARM\Pack\ARM\mbedTLS\1.0.0\library\" */
  const char *pos = strrchr (file, '\\');
  if (pos != NULL) file = pos + 1;
#endif
  fprintf ((FILE *)ctx, "%s:%04d: %s", file, line, str);
  fflush ((FILE *)ctx);
}
#endif

/**
  \brief       Initialize TLS interface.
  \return      none.
*/
void netTLS_init (void) {
  memset (&sec_scb, 0, sizeof (sec_scb));
  memset (&ssl, 0, sizeof (ssl));
  memset (&ctrl, 0, sizeof (ctrl));
  ctrl.thread_id = OS_THREAD_NEW (netTLS_Thread);
}

/**
  \brief       Allocate TLS socket and assign a TCP socket.
  \param[in]   cb_func  event listening TCP callback function.
  \return      TCP socket handle number or 0= no socket assigned.
*/
int32_t netTLS_socket (netTCP_cb_t cb_func) {
  TLS_INFO *sec_s;
  int32_t i,sock;

  for (i = 0, sec_s = &sec_scb[0]; i < TLS_NUM_SOCKS; sec_s++, i++) {
    if (sec_s->State == TLS_STATE_FREE) {
      /* Get TCP socket first */
      sock = net_tcp_get_socket (cb_tls);
      if (sock <= 0) {
        break;
      }
      sec_s->Socket  = (uint8_t)sock;
      sec_s->ssl     = &ssl[i];
      sec_s->cb_func = cb_func;
      sec_s->State   = TLS_STATE_IDLE;
      return (sock);
    }
  }
  return (0);
}

/**
  \brief       Allocate memory for TLS send buffer.
  \param[in]   size  number of bytes to allocate.
  \return      pointer to allocated memory.
*/
uint8_t *netTLS_buf (uint32_t size) {
  NET_BUFFER *netbuf;

  netbuf = __BUFFER(net_mem_alloc (QUEUE_HEADER_LEN + size));
  if (netbuf == NULL) {
    /* This is for safety, actually not used */
    return (sec_buf);
  }
  /* Marker used for invalid buffer check */
  netbuf->next = netbuf;
  return (&netbuf->data[0]);
}

/**
  \brief       Close TLS socket request.
  \param[in]   sock  socket to close.
  \return      none.
*/
void netTLS_close (int32_t sock) {
  TLS_INFO *sec_s;

  /* Find the socket owner */
  sec_s = tls_map_socket (sock);
  if (sec_s == NULL) {
    return;
  }
  sec_s->CloseRq = true;
  thread_wakeup ();
}

/**
  \brief       Send data on connected TLS socket.
  \param[in]   sock  socket descriptor.
  \param[in]   buf   pointer to data buffer.
  \param[in]   len   length of data (in bytes).
  \return      none.
*/
void netTLS_send (int32_t sock, const uint8_t *buf, uint32_t len) {
  TLS_INFO *sec_s;
  NET_BUFFER *netbuf;

  if ((buf == NULL) || (buf == sec_buf)) {
    return;
  }
  netbuf = __BUFFER(buf - NET_BUFFER_HLEN);
  /* Check the marker */
  if (netbuf->next != netbuf) {
    /* Invalid buffer, ignore it */
    return;
  }
  /* Find the socket owner */
  sec_s = tls_map_socket (sock);
  if (sec_s == NULL) {
    net_mem_free ((NET_FRAME *)netbuf);
    return;
  }
  if ((sec_s->t_buf != NULL) || (sec_s->State != TLS_STATE_APPDATA)) {
    /* Previous frame still pending or invalid state */
    /* This is unexpected, should never happen       */
    net_mem_free ((NET_FRAME *)netbuf);
    return;
  }
  netbuf->next   = NULL;
  netbuf->length = (uint16_t)len;
  netbuf->index  = 0;
  sec_s->t_buf   = netbuf;
  thread_wakeup ();
}

/**
  \brief       TLS helper thread.
  \param[in]   arg  dummy parameter.
  \return      none.
*/
static void netTLS_Thread (void const *arg) {
  int32_t ret;

  (void)arg;
  ret = tls_init ();
  if (ret != 0) {
    /* TLS layer failed to initialize, reasons: */
    /*   - heap size too small                  */
    /*   - server certificate or key error      */
    /*   - random number generator problem      */
    /* Check MBEDTLS_ERR_x code for the reason. */
    ctrl.ssl_error = ret;
    ctrl.init_done = false;
  }
  else {
    /* Initialize successful */
    ctrl.init_done = true;
  }
  while (1) {
    OS_FLAG_WAIT (0x0001);
    do {
      ctrl.running = true;
      tls_run ();
      ctrl.running = false;
    } while (ctrl.busy);
  }
}

/**
  \brief       Wake-up TLS helper thread.
  \return      none.
*/
static void thread_wakeup (void) {
  if (ctrl.running) {
    /* Thread currently running */
    ctrl.busy = true;
    return;
  }
  OS_FLAG_SET (0x0001);
}

/**
  \brief       Initialize TLS socket layer.
  \return      TLS error code (0= no error).
*/
static int32_t tls_init (void) {
  int32_t ret;

  mbedtls_ssl_config_init (&conf);
  mbedtls_x509_crt_init (&srvcert);
  mbedtls_pk_init (&pkey);
  mbedtls_ctr_drbg_init (&ctr_drbg);
  mbedtls_entropy_init (&entropy);
#if defined(MBEDTLS_SSL_CACHE_C)
  mbedtls_ssl_cache_init (&cache);
#endif

#if defined(MBEDTLS_DEBUG_C)
  mbedtls_debug_set_threshold (DEBUG_LEVEL);
  mbedtls_ssl_conf_dbg (&conf, out_debug, stdout);
#endif

  /* 1. Load the certificates and private key */

  ret = mbedtls_x509_crt_parse (&srvcert, NetSecurity_ServerCert,
                                          NetSecurity_ServerCert_Len);
  if (ret != 0) {
    /* Server certificate parse failed */
    return (ret);
  }

  ret = mbedtls_x509_crt_parse (&srvcert, NetSecurity_ServerCA,
                                          NetSecurity_ServerCA_Len);
  if (ret != 0) {
    /* Server CA certificate(s) parse failed */
    return (ret);
  }

  ret = mbedtls_pk_parse_key (&pkey, NetSecurity_ServerKey,
                                     NetSecurity_ServerKey_Len, NULL, 0);
  if (ret != 0) {
    /* Server key parse failed */
    return (ret);
  }

  /* 2. Seed the RNG */
  ret = mbedtls_ctr_drbg_seed (&ctr_drbg, mbedtls_entropy_func, &entropy,
                               (const uint8_t *)pers, strlen (pers));
  if (ret != 0) {
    /* Random number generator seed failed */
    return (ret);
  }

  /* 3. Setup stuff */
  ret = mbedtls_ssl_config_defaults (&conf, MBEDTLS_SSL_IS_SERVER,
                                            MBEDTLS_SSL_TRANSPORT_STREAM,
                                            MBEDTLS_SSL_PRESET_DEFAULT);
  if (ret != 0) {
    /* Configure defaults failed */
    return (ret);
  }

  mbedtls_ssl_conf_rng (&conf, mbedtls_ctr_drbg_random, &ctr_drbg);

#if defined(MBEDTLS_SSL_CACHE_C)
  mbedtls_ssl_conf_session_cache (&conf, &cache,
                                  mbedtls_ssl_cache_get,
                                  mbedtls_ssl_cache_set);
#endif
  mbedtls_ssl_conf_ca_chain (&conf, srvcert.next, NULL);
  ret = mbedtls_ssl_conf_own_cert (&conf, &srvcert, &pkey);
  if (ret != 0) {
    /* Set own certificate chain and private key failed */
    return (ret);
  }
  return (0);
}

/**
  \brief       Run TLS service.
  \return      none.
  \note        Generates native TCP callbacks for the service.
*/
static void tls_run (void) {
  TLS_INFO *sec_s;
  NET_BUFFER *tbuf;
  int32_t i,ret;

  ctrl.busy = false;
  for (i = 0, sec_s = &sec_scb[0]; i < TLS_NUM_SOCKS; sec_s++, i++) {
    switch (sec_s->State) {
      case TLS_STATE_FREE:
      case TLS_STATE_IDLE:
        break;

      case TLS_STATE_SETUP:
        /* Socket is connected, now start SSL handshaking */
        sec_s->CloseRq   = false;
        sec_s->ssl_error = 0;
        ret = mbedtls_ssl_setup (sec_s->ssl, &conf);
        if (ret != 0) {
          /* Setup failed, probably out of memory */
          sec_s->ssl_error = ret;
          goto close;
        }
        mbedtls_ssl_set_bio (sec_s->ssl, sec_s, bio_send, bio_recv, NULL);
        sec_s->State = TLS_STATE_HANDSHAKE;
        ctrl.busy    = true;
        break;

      case TLS_STATE_HANDSHAKE:
        ret = mbedtls_ssl_handshake (sec_s->ssl);
        if ((ret == MBEDTLS_ERR_SSL_WANT_READ) ||
            (ret == MBEDTLS_ERR_SSL_WANT_WRITE)) {
          break;
        }
        if (ret != 0) {
          /* Handshake error, close and reset session */
          sec_s->ssl_error = ret;
          goto close;
        }
        net_sys_lock ();
        sec_s->cb_func (sec_s->Socket, netTCP_EventEstablished, (NET_ADDR *)&sec_s->Peer, NULL, 0);
        net_sys_unlock ();
        sec_s->State = TLS_STATE_APPDATA;
        ctrl.busy    = true;
        break;

      case TLS_STATE_APPDATA:
        /* Send and receive encrypted data */
        if (sec_s->t_buf != NULL) {
          /* Data available for sending */
          tbuf = sec_s->t_buf;
          ret  = mbedtls_ssl_write (sec_s->ssl,
                                    &tbuf->data[tbuf->index], tbuf->length);
          if (ret > 0) {
            /* Some data sent successfully */
            tbuf->index += ret;
            if (tbuf->index < tbuf->length) {
              /* Buffer not completed, try again later */
              if (sec_s->CloseRq) {
                break;
              }
            }
            else {
              /* Send completed, release buffer */
              free_tx_buf (sec_s);
              if (sec_s->CloseRq) {
                goto close;
              }
              net_sys_lock ();
              sec_s->cb_func (sec_s->Socket, netTCP_EventACK, (NET_ADDR *)&sec_s->Peer, NULL, 0);
              net_sys_unlock ();
            }
          }
          else if ((ret != MBEDTLS_ERR_SSL_WANT_READ) &&
                   (ret != MBEDTLS_ERR_SSL_WANT_WRITE)) {
            /* Error, close session */
            sec_s->ssl_error = ret;
            goto close;
          }
        }
        if (sec_s->CloseRq) {
          goto close;
        }
        ret = mbedtls_ssl_read (sec_s->ssl, sec_buf, sizeof(sec_buf));
        if (ret > 0) {
          /* Data received, send a notification */
          net_sys_lock ();
          sec_s->cb_func (sec_s->Socket, netTCP_EventData, (NET_ADDR *)&sec_s->Peer, sec_buf, (uint32_t)ret);
          net_sys_unlock ();
          ctrl.busy = true;
        }
        else if ((ret != MBEDTLS_ERR_SSL_WANT_READ) &&
                 (ret != MBEDTLS_ERR_SSL_WANT_WRITE)) {
          /* Error, close session */
          if (ret != MBEDTLS_ERR_NET_CONN_RESET) {
            sec_s->ssl_error = ret;
          }
close:    sec_s->State = TLS_STATE_CLOSE;
          ctrl.busy    = true;
        }
        break;

      case TLS_STATE_CLOSE:
        /* Socket has been closed */
        if (!sec_s->ssl_error) netTCP_Close (sec_s->Socket);
        else                   netTCP_Abort (sec_s->Socket);
        net_sys_lock ();
        if (!sec_s->CloseRq) {
          /* Application is not aware of socket closing */
          sec_s->cb_func (sec_s->Socket, netTCP_EventClosed, (NET_ADDR *)&sec_s->Peer, NULL, 0);
        }
        queue_release (sec_s);
        net_sys_unlock ();
        free_tx_buf (sec_s);
        mbedtls_ssl_free (sec_s->ssl);
        sec_s->CloseRq = false;
        sec_s->State   = TLS_STATE_IDLE;
        break;
    }
  }
}

/**
  \brief       Map TCP socket to existing TLS session.
  \param[in]   socket  TCP socket handle.
  \return      pointer to TLS session control block.
*/
static TLS_INFO *tls_map_socket (int32_t socket) {
  TLS_INFO *sec_s;
  int32_t i;

  if (!ctrl.init_done) {
    /* TLS layer not yet initialized */
    return (NULL);
  }
  for (i = 0, sec_s = &sec_scb[0]; i < TLS_NUM_SOCKS; sec_s++, i++) {
    if (sec_s->Socket == (uint8_t)socket) {
      return (sec_s);
    }
  }
  return (NULL);
}

/**
  \brief       Socket TCP event callback notification.
  \param[in]   socket  native TCP socket handle.
  \param[in]   event   notification event.
  \param[in]   addr    structure containing remote IP address and port.
  \param[in]   buf     pointer to a buffer containing the data.
  \param[in]   len     length of the data.
  \return      status:
               - true on success,
               - false on error.
  \note        This function is called from protected environment.
               No need to call net_sys_lock/net_sys_unlock.
*/
static uint32_t cb_tls (int32_t socket, netTCP_Event event, const NET_ADDR *addr,
                                        const uint8_t *buf, uint32_t len) {
  TLS_INFO *sec_s;
  NET_BUFFER *netbuf;

  /* Find the socket owner */
  sec_s = tls_map_socket (socket);
  if (sec_s == NULL) {
    return (false);
  }
  switch (event) {
    case netTCP_EventConnect:
      if (sec_s->State > TLS_STATE_IDLE) {
        /* Reject new connection if SSL state not idle yet */
        return (false);
      }
      net_addr_copy (&sec_s->Peer, (const __ADDR *)addr);
      /* Let the service handle connection accept/deny */
      return (sec_s->cb_func (sec_s->Socket, event, addr, NULL, 0));

    case netTCP_EventAborted:
    case netTCP_EventClosed:
      /* Connection closed, virtual netTCP_EventClose will */
      /* be generated to the application from TLS thread.  */
      thread_wakeup ();
      break;

    case netTCP_EventEstablished:
      /* Connected, handle TLS handshaking in this module.  */
      /* Generate virtual 'netTCP_EventEstablished' to the  */
      /* application when the handshaking is over.          */
      sec_s->State = TLS_STATE_SETUP;
      thread_wakeup ();
      break;

    case netTCP_EventACK:
      /* Data has been acknowledged */
      thread_wakeup ();
      break;

    case netTCP_EventData:
      /* Data has been received, store the data to receive queue. */
      netbuf = __BUFFER(net_mem_alloc (QUEUE_HEADER_LEN + len));
      if (netbuf == NULL) {
        /* No memory, do not ack this packet */
        return (NET_TCP_NO_ACK);
      }
      netbuf->length = (uint16_t)len;
      netbuf->index  = 0;
      memcpy (&netbuf->data[0], buf, len);
      queue_add_buf (sec_s, netbuf);
      thread_wakeup ();
      break;
  }
  return (true);
}

/**
  \brief       Read data from TCP socket.
  \param[in]   ctx  TLS session descriptor.
  \param[out]  buf  pointer to application data buffer to store the data to.
  \param[in]   len  size of application data buffer (in bytes).
  \return      number of bytes read or status information (retv < 0).
  \note        Reads data that is buffered for the session.
*/
static int32_t bio_recv (void *ctx, unsigned char *buf, uint32_t len) {
  TLS_INFO *sec_s = ctx;
  NET_BUFFER *netbuf;
  uint32_t nr,dlen;

  if (sec_s->r_queue == NULL) {
    /* No data available */
    if (netTCP_GetState (sec_s->Socket) != netTCP_StateESTABLISHED) {
      return (MBEDTLS_ERR_NET_CONN_RESET);
    }
    return (MBEDTLS_ERR_SSL_WANT_READ);
  }

  for (nr = 0; (nr < len) && sec_s->r_queue; ) {
    netbuf = sec_s->r_queue;
    dlen   = netbuf->length - netbuf->index;
    if (dlen > (len - nr)) {
      dlen = len - nr;
    }
    memcpy (&buf[nr], &netbuf->data[netbuf->index], dlen);
    nr += dlen;
    if ((netbuf->index + dlen) < netbuf->length) {
      /* User buffer is full, data still remaining in this block */
      netbuf->index += dlen;
      break;
    }
    /* This buffer done, get next from the queue */
    sec_s->r_queue = netbuf->next;
    net_mem_free ((NET_FRAME *)netbuf);
  }
  return ((int32_t)nr);
}

/**
  \brief       Send data to TCP socket.
  \param[in]   ctx  TLS session descriptor.
  \param[in]   buf  pointer to application data buffer to transmit.
  \param[in]   len  length of data (in bytes).
  \return      number of bytes sent or status information (retv < 0).
*/
static int32_t bio_send (void *ctx, const uint8_t *buf, uint32_t len) {
  TLS_INFO *sec_s = ctx;
  uint32_t maxsize;
  uint8_t *tcp_buf;
  netStatus retv;

  if (!netTCP_SendReady (sec_s->Socket)) {
    if (netTCP_GetState (sec_s->Socket) != netTCP_StateESTABLISHED) {
      /* Socket is not connected */
      return (MBEDTLS_ERR_NET_CONN_RESET);
    }
    /* Socket is waiting for ACK */
    return (MBEDTLS_ERR_SSL_WANT_WRITE);
  }
  maxsize = netTCP_GetMaxSegmentSize (sec_s->Socket);
  if (len > maxsize) {
    len = maxsize;
  }
  tcp_buf = netTCP_GetBuffer (len | 0x80000000);
  if (tcp_buf == NULL) {
    /* No memory, try again later */
    return (MBEDTLS_ERR_SSL_WANT_WRITE);
  }
  memcpy (tcp_buf, buf, len);
  retv = netTCP_Send (sec_s->Socket, tcp_buf, len);
  if (retv == netOK) {
    /* Return number of bytes sent */
    return ((int32_t)len);
  }
  if (retv == netBusy) {
    return (MBEDTLS_ERR_SSL_WANT_WRITE);
  }
  if (retv == netInvalidParameter) {
    return (MBEDTLS_ERR_NET_INVALID_CONTEXT);
  }
  if (retv == netWrongState) {
    return (MBEDTLS_ERR_NET_CONN_RESET);
  }
  return (MBEDTLS_ERR_NET_SEND_FAILED);
}

/**
  \brief       Add data buffer to socket receive queue tail.
  \param[in]   sec_s   TLS session descriptor.
  \param[in]   netbuf  pointer to netbuf containing the data.
  \return      none.
*/
static void queue_add_buf (TLS_INFO *sec_s, NET_BUFFER *netbuf) {
  NET_BUFFER *next_buf;

  netbuf->next = NULL;
  if (sec_s->r_queue == NULL) {
    /* First data block added */
    sec_s->r_queue = netbuf;
  }
  else {
    /* Scan the queue chain */
    next_buf = sec_s->r_queue;
    while (next_buf->next != NULL) {
      next_buf = next_buf->next;
    }
    /* Append the buffer to queue tail */
    next_buf->next = netbuf;
  }
}

/**
  \brief       Release chained netbuf data buffers.
  \param[in]   sec_s  TLS session descriptor.
  \return      none.
*/
static void queue_release (TLS_INFO *sec_s) {
  NET_BUFFER *netbuf, *next_buf;

  netbuf = sec_s->r_queue;
  sec_s->r_queue = NULL;
  for ( ; netbuf; netbuf = next_buf) {
    next_buf = netbuf->next;
    net_mem_free ((NET_FRAME *)netbuf);
  }
}

/**
  \brief       Release TX data buffer.
  \param[in]   sec_s  TLS session descriptor.
  \return      none.
*/
static void free_tx_buf (TLS_INFO *sec_s) {
  if (sec_s->t_buf) {
    net_mem_free ((NET_FRAME *)sec_s->t_buf);
    sec_s->t_buf = NULL;
  }
}
